// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       binary_generator.h
 *  \brief      メッセージデータのバイナリ生成クラス
 *  \date       Since: June 28, 2021. 2:04:29 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_BINARY_GENERATOR_H_1624813469
#define INCGUARD_BINARY_GENERATOR_H_1624813469

#include "indexed_character.h"

#include <cstdint>
#include <memory>
#include <string>
#include <vector>

#include <mtk/text_format_argument.h>
#include <mtk/text_hash.h>

#include "run_options.h"
#include "text_replacer.h"

namespace msgconv
{
class BinaryGenerator : public TextReplaceCommandProcessor
{
public:
    BinaryGenerator() noexcept;

    bool Generate(const RunOptions::Options &options) noexcept;

    void PrintCSVInfo() const noexcept;

private:
    // バイナリデータのフラグ情報
    static constexpr const uint32_t kBinaryFlagIndexed = 1 << 0;   //!< インデックス文字列かのフラグ

    bool LoadMetrics(const std::filesystem::path &fontMetricsPath) noexcept;
    bool LoadCSV(const std::filesystem::path &csvPath) noexcept;

    bool BuildMessagePool() noexcept;
    bool BuildIndexedMessagePool() noexcept;
    bool BuildOffsetPool() noexcept;

    bool SaveBinaryFile(const std::filesystem::path &binaryFilePath) const noexcept;
    bool SaveMessageID(const std::filesystem::path &outputPath) const noexcept;
    bool SaveLanguageID(const std::filesystem::path &outputPath) const noexcept;

    virtual bool DoReplaceCommand(FILE *fp, const TextReplacer::Node &node, const std::filesystem::path &srcFilePath) noexcept override;

    IndexedCharacter ToIndexedCharacter(char32_t character) const noexcept;

    std::vector<IndexedCharacter> ToIndexedCharactersFromFormatOption(const MTK::FormatArgument::Option &option) noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      テキストからハッシュ値を取得
     *  \param[in]  text    テキスト
     *  \return     ハッシュ値
     */
    /* ------------------------------------------------------------------------- */
    constexpr uint32_t GetHash(const char *text) const noexcept
    {
        return MTK::Hash::FNV1a(text, _hashSeed);
    }

    uint32_t _hashSeed;

    //! フォント情報
    struct FontInfo
    {
        uint32_t identify;              //!< 識別子
        uint32_t revision;              //!< リビジョン情報
        uint32_t characterCount;        //!< 文字数
        uint32_t size;                  //!< フォントサイズ（コンバータへの指定値）
        uint32_t maxHeight;             //!< 高さの最大サイズ
        uint32_t texturePageCount;      //!< テクスチャのページ数
        uint32_t alignmentSize;         //!< アライメントサイズ
        uint32_t flags;                 //!< フラグ情報
    };
    FontInfo _fontInfo;

    //! 文字の情報
    struct CharacterInfo
    {
        char32_t code;          //!< 文字コード（UTF-32）
        uint32_t width;         //!< 幅
        uint32_t height;        //!< 高さ
        uint32_t advance;       //!< 送り幅
        uint32_t bearingX;      //!< X方向のベアリング
        uint32_t bearingY;      //!< Y方向のベアリング
        uint32_t textureX;      //!< テクスチャのX座標
        uint32_t textureY;      //!< テクスチャのY座標
        uint32_t textureWidth;  //!< テクスチャ上の幅
        uint32_t textureHeight; //!< テクスチャ上の高さ
        uint32_t texturePage;   //!< テクスチャのページ
    };
    std::unique_ptr<CharacterInfo[]> _characterArray;

    //! 言語情報
    struct LanguageInfo
    {
        std::string name;       //!< 言語名
        uint32_t hash;          //!< 言語名のハッシュ値
        int fieldIndex;         //!< CSVのフィールドインデックス
    };
    std::vector<LanguageInfo> _languages;

    //! メッセージID
    struct MessageID
    {
        std::string name;       //!< ID名
        uint32_t hash;          //!< ID名のハッシュ値
        int recordIndex;        //!< CSVのレコードインデックス
    };
    std::vector<MessageID> _identifiers;

    //! メッセージデータ
    struct MessageData
    {
        std::string text;       //!< テキスト (UTF-8)
        size_t offset;          //!< メッセージプール上のオフセット
    };
    std::vector<std::vector<MessageData>> _messageTable;

    //! メッセージプール
    struct MessagePool
    {
        bool isIndexed;                         //!< インデックス文字列を格納しているかのフラグ
        std::unique_ptr<std::byte[]> buffer;    //!< メッセージプールのバッファ
        size_t size;                            //!< メッセージプールのバッファサイズ
    };
    MessagePool _messagePool;

    //! オフセットプール
    struct OffsetPool
    {
        std::unique_ptr<uint32_t[]> buffer;     //!< オフセットプールのバッファ
        size_t elementCount;                    //!< 要素数
    };
    OffsetPool _offsetPool;
};
}   // namespace msgconv

#endif	// INCGUARD_BINARY_GENERATOR_H_1624813469

// vim: et ts=4 sw=4 sts=4
