// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       message.cc
 *  \brief      表示メッセージ
 *  \date       Since: February 8, 2023. 22:28:41 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include "message.h"

#include <mtk/text_hash.h>
#include <mtk/environment.h>

#if defined(MTK_TARGET_WINDOWS)
#include <mbctype.h>
#include <Windows.h>
#undef GetMessage
#include <mtk/text_converter.h>
#endif

namespace msgconv
{
namespace
{
// 現在の言語ID
static LanguageID sCurrentLanguageID = LanguageID::en_US;
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      メッセージを初期化
 */
/* ------------------------------------------------------------------------- */
void InitMessage() noexcept
{
#if defined(MTK_TARGET_WINDOWS)
    // WindowsではUTF-8 or CP932かつ日本語環境で日本語メッセージを表示
    if (PRIMARYLANGID(GetUserDefaultLangID()) == LANG_JAPANESE)
    {
        if (auto mbcp = _getmbcp(); (mbcp == CP_UTF8) || (mbcp == 932))
        {
            sCurrentLanguageID = LanguageID::ja_JP;
        }
    }
#else
    // 環境変数"LANG"を読み取り
    if (auto lang = MTK::Environment::GetValue("LANG"); !lang.empty())
    {
        switch (MTK::Hash::FNV1a(lang.c_str()))
        {
            // 日本語UTF-8
            case MTK::Hash::FNV1a("ja_JP.UTF-8"):
                sCurrentLanguageID = LanguageID::ja_JP;
                break;

            default:
                break;
        }
    }
#endif
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      出力メッセージを取得
 *  \param[in]  msgID   メッセージID
 *  \param[in]  args    フォーマット引数
 *  \return     出力メッセージ
 */
/* ------------------------------------------------------------------------- */
std::string GetOutputMessage(MsgID msgID, const MTK::FormatArgs &args) noexcept
{
#if defined(MTK_TARGET_WINDOWS)
    // WindowsではCP932の場合にメッセージをShift-JISに変換する
    std::string outputMessage;
    if (_getmbcp() == 932)
    {
        auto message = MTK::Format(GetMessage(sCurrentLanguageID, msgID), args);
        if (!MTK::ToShiftJIS(outputMessage, message.c_str(), message.size(), MTK::TextEncoding::UTF8))
        {
            outputMessage = MTK::Format(GetMessage(LanguageID::en_US, msgID), args);
        }
    }
    else
    {
        outputMessage = MTK::Format(GetMessage(sCurrentLanguageID, msgID), args);
    }

    return outputMessage;
#else
    // Windows以外ではそのまま
    return MTK::Format(GetMessage(sCurrentLanguageID, msgID), args);
#endif
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      メッセージの表示
 *  \param[in]  msgID   メッセージID
 *  \param[in]  args    フォーマット引数
 */
/* ------------------------------------------------------------------------- */
void PutMessage(MsgID msgID, const MTK::FormatArgs &args) noexcept
{
    puts(GetOutputMessage(msgID, args).c_str());
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      エラーメッセージの表示
 *  \param[in]  msgID   メッセージID
 *  \param[in]  args    フォーマット引数
 */
/* ------------------------------------------------------------------------- */
void PutError(MsgID msgID, const MTK::FormatArgs &args) noexcept
{
    fprintf(stderr, "Error: %s\n", GetOutputMessage(msgID, args).c_str());
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      警告メッセージの表示
 *  \param[in]  msgID   メッセージID
 *  \param[in]  args    フォーマット引数
 */
/* ------------------------------------------------------------------------- */
void PutWarning(MsgID msgID, const MTK::FormatArgs &args) noexcept
{
    fprintf(stderr, "Warning: %s\n", GetOutputMessage(msgID, args).c_str());
}
}   // namespace msgconv

// vim: et ts=4 sw=4 sts=4
