// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       text_converter.cc
 *  \brief      テキストコンバータ
 *  \date       Since: February 24, 2023. 2:57:38 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mtk/text_converter.h>

namespace MTK
{
namespace
{
#if defined(MTK_TARGET_WINDOWS)
// Unicode - ShiftJIS 変換表
constexpr const struct SJISCharacterInfo
{
    char16_t unicode;   // UTF-16のコード
    uint16_t sjis;      // Shift-JISのコード
}
kShiftJISTable[] =
{
    #include "sjis_table.inc"
};

// 変換表のサイズ
constexpr size_t kShiftJISTableSize = sizeof(kShiftJISTable) / sizeof(SJISCharacterInfo);
#endif
}

#if defined(MTK_TARGET_WINDOWS)
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      UnicodeをShiftJISに変換
 *  \param[out] dest        変換結果の格納先
 *  \param[in]  src         変換元の先頭アドレス
 *  \param[in]  size        変換元のサイズ
 *  \param[in]  encoding    変換元のエンコードタイプ
 *  \param[in]  existBOM    変換元にBOMが付与されているかのフラグ（TextEncoding::Autoの場合のみ参照）
 *  \note
 *      Windows以外の環境ではShift-JISを使うべきではないのでWindowsでのみ利用可能．
 *      有効にするとコードサイズが27KiBくらい増える．
 */
/* ------------------------------------------------------------------------- */
bool ToShiftJIS(std::string &dest, const void *src, size_t size, TextEncoding encoding, bool existBOM) noexcept
{
    dest.clear();
    dest.reserve(4096);

    TextStream stream(reinterpret_cast<const uint8_t *>(src), size, encoding, existBOM);

    while (!stream.IsTerminated())
    {
        auto character = stream.Read();
        if (!character.IsValid())
        {
            dest.clear();
            return false;
        }

        // 0xFF以下はそのまま
        if (auto code = character.GetCode(); code < 0xFF)
        {
            dest += static_cast<char>(code);
        }
        else
        {
            size_t min = 0;
            size_t max = kShiftJISTableSize;
            bool isFounded = false;
            
            // バイナリサーチ
            do
            {
                size_t index = min + (max - min) / 2;
                const auto &charInfo = kShiftJISTable[index];
                
                if (charInfo.unicode == code)
                {
                    // 見つかったら対応するSJISコードを格納先に追加
                    dest += static_cast<char>((charInfo.sjis >> 8) & 0xFF);
                    dest += static_cast<char>(charInfo.sjis & 0xFF);
                    isFounded = true;
                    break;
                }
                else if (charInfo.unicode > code)
                {
                    max = index;
                }
                else if (charInfo.unicode < code)
                {
                    min = index + 1;
                }
                
            } while(min != max);

            // 見つからなかったらとりあえず "??" を追加しておく
            if (!isFounded)
            {
                dest += "??";
            }
        }
    }
    
    return true;
}
#endif

}   // namespace MTK

// vim: et ts=4 sw=4 sts=4
